package com.flyfish.framework.logging.service;

import com.alibaba.fastjson.JSON;
import com.flyfish.framework.annotations.Operation;
import com.flyfish.framework.beans.meta.RestBean;
import com.flyfish.framework.context.UserContext;
import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.logging.domain.DomainDescriptions;
import com.flyfish.framework.logging.domain.Log;
import com.flyfish.framework.utils.ReflectionUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.Signature;
import org.aspectj.lang.reflect.MethodSignature;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.data.util.CastUtils;
import org.springframework.stereotype.Service;
import org.springframework.web.bind.annotation.RequestMapping;

import java.util.Optional;

/**
 * 日志管理器
 *
 * @author wangyu
 */
@Service
@Slf4j
public class LogManager {

    /**
     * 尝试记录日志
     *
     * @param joinPoint 接入点
     * @param e         异常
     */
    public void tryLog(ProceedingJoinPoint joinPoint, Object result, Throwable e) {
        Signature signature = joinPoint.getSignature();
        if (signature instanceof MethodSignature) {
            MethodSignature methodSignature = (MethodSignature) signature;
            Operation operation = AnnotationUtils.findAnnotation(methodSignature.getMethod(), Operation.class);
            if (null != operation) {
                String module = operation.module();
                Class<?> targetClass = joinPoint.getTarget().getClass();
                if (StringUtils.isBlank(module)) {
                    module = ReflectionUtils.getGenericType(targetClass)
                            .flatMap(clazz -> {
                                DomainDescriptions description = DomainDescriptions.getByClass(CastUtils.cast(clazz));
                                if (null != description) {
                                    return Optional.of(description.getName());
                                }
                                return Optional.ofNullable(AnnotationUtils.findAnnotation(clazz, RestBean.class))
                                        .map(RestBean::name);
                            })
                            .orElse("未知模块");
                }
                log.info("监测到支持的切点：{}, {}", module, operation.value());
                // 构建日志
                Log log = new Log();
                log.setSuccess(null == e);
                log.setBody(bodyString(joinPoint.getArgs()));
                log.setModule(module);
                log.setBusiness(operation.value());
                log.setError(Optional.ofNullable(e).map(JSON::toJSONString).orElse(null));
                log.setResponse(Optional.ofNullable(result).map(JSON::toJSONString).orElse(null));
                log.setOperator(UserContext.sharedContext().map(UserContext::currentUser).map(Domain::getName).orElse("未知"));
                log.setSignature(joinPoint.getSignature().getName());
                log.setUri(Optional.ofNullable(AnnotationUtils.findAnnotation(targetClass, RequestMapping.class))
                        .map(mapping -> mapping.value())
                        .filter(ArrayUtils::isNotEmpty)
                        .map(array -> array[0])
                );
                // 写入日志
                logService.create()
            }
        }
    }

    private String bodyString(Object[] args) {
        if (ArrayUtils.isEmpty(args)) {
            return null;
        }
        if (args.length == 1) {
            return JSON.toJSONString(args[0]);
        }
        return JSON.toJSONString(args);
    }
}
